module DM
module ProfileBuilder2
module Format
##################
def self.length(num)
num=num.to_f
uo=Sketchup.active_model.options["UnitsOptions"]
unit=uo["LengthUnit"]
precision=uo["LengthPrecision"]
precision=3
case unit
when 0,1 #inches
	factor=(1.0/12.0)
when 2,3,4 #milliimeters
	factor=(0.0254)
end
length="%.#{precision}f" %(num*factor)
return length
end
##################
def self.small_length(num)
num=num.to_f
uo=Sketchup.active_model.options["UnitsOptions"]
unit=uo["LengthUnit"]
precision=uo["LengthPrecision"]
precision=3
case unit
when 0,1 #inches
	factor=(1.0)
when 2,3,4 #milliimeters
	factor=(25.4)
end
length="%.#{precision}f" %(num*factor)
return length
end
#############
##################
def self.area(num)
num=num.to_f
uo=Sketchup.active_model.options["UnitsOptions"]
unit=uo["LengthUnit"]
precision=uo["LengthPrecision"]
precision=3
case unit
when 0,1 #inches
	factor=(1.0/12.0)**2
when 2, 3, 4 #milliimeters
	factor=(0.0254)**2
end
area="%.#{precision}f" %(num*factor)
return area
end
##
def self.cost(num,pre_symbol='', thousands=',', decimal='.', post_symbol=nil)
	if num
		"#{pre_symbol}#{( "%.2f" % num ).gsub(/(\d)(?=(?:\d{3})+(?:$|\.))/,"\\1#{thousands}")}#{post_symbol}"
	else
		"0.0"
	end
end
##################
def self.volume(num)
num=num.to_f
uo=Sketchup.active_model.options["UnitsOptions"]
unit=uo["LengthUnit"]
precision=uo["LengthPrecision"]
precision=3
suppress_units=uo["SuppressUnitsDisplay"]
case unit
when 0,1 #inches or feet
	factor=(1.0/12.0)**3
when 2,3,4 #milliimeters, centimeters, or meters
	factor=(0.0254)**3
end
vol="%.#{precision}f" %(num*factor)
return vol
end
##################
def self.weight(num)
num=num.to_f
uo=Sketchup.active_model.options["UnitsOptions"]
unit=uo["LengthUnit"]
precision=uo["LengthPrecision"]
suppress_units=uo["SuppressUnitsDisplay"]
precision=3
case unit
when 0,1 #inches or feet
	factor=1.0
when 2,3,4 #milliimeters, centimeters, or meters
	factor=0.453592
end
vol="%.#{precision}f" %(num*factor)
return vol
end
#####
def self.volume_unit(html=true)
uo=Sketchup.active_model.options["UnitsOptions"]
unit=uo["LengthUnit"]
precision=uo["LengthPrecision"]
suppress_units=uo["SuppressUnitsDisplay"]
case unit
when 0,1 #inches, feet
	sym="ft<sup>3</sup>" if html
	sym="ft3" if not html
when 2, 3, 4 #milliimeters
	sym="m<sup>3</sup>" if html
	sym="m3" if not html
end
return sym
end
#####
def self.weight_unit
uo=Sketchup.active_model.options["UnitsOptions"]
unit=uo["LengthUnit"]
precision=uo["LengthPrecision"]
case unit
when 0,1 #inches, feet
	sym="lbs"
when 2, 3, 4 #milliimeters
	sym="kg"
end
return sym
end
#####
def self.length_unit
uo=Sketchup.active_model.options["UnitsOptions"]
unit=uo["LengthUnit"]
precision=uo["LengthPrecision"]
suppress_units=uo["SuppressUnitsDisplay"]
case unit
when 0,1 #inches
	sym="ft"
when 2,3,4 #milliimeters
	sym="m"
end
#sym="" if suppress_units==true
return sym
end
#####
def self.small_length_unit
uo=Sketchup.active_model.options["UnitsOptions"]
unit=uo["LengthUnit"]
precision=uo["LengthPrecision"]
suppress_units=uo["SuppressUnitsDisplay"]
case unit
when 0,1 #inches
	sym="in"
when 2,3,4 #milliimeters
	sym="mm"
end
return sym
end
##
#####
def self.area_unit(html=true)
uo=Sketchup.active_model.options["UnitsOptions"]
unit=uo["LengthUnit"]
precision=uo["LengthPrecision"]
suppress_units=uo["SuppressUnitsDisplay"]
case unit
when 0,1 #inches, feet
	sym="ft<sup>2</sup>" if html
	sym="ft2" if not html
when 2, 3, 4 
	sym="m<sup>2</sup>" if html
	sym="m2" if not html
end
#sym="" if suppress_units==true
return sym
end
#####
def self.default_density_unit(html=true)
uo=Sketchup.active_model.options["UnitsOptions"]
unit=uo["LengthUnit"]
case unit
when 0,1 #inches, feet
	sym="lb/ft<sup>3</sup>" if html
	sym="lb/ft3" if not html
when 2, 3, 4 
	sym="kg/m<sup>3</sup>" if html
	sym="kg/m3" if not html
end
return sym
end
###
def self.default_cost_unit
uo=Sketchup.active_model.options["UnitsOptions"]
unit=uo["LengthUnit"]
case unit
when 0,1 #inches, feet
	sym="lb"
when 2, 3, 4 
	sym="kg"
end
return sym
end
end #module format

class QMember
@@dict="PBMTO"
public
##
def initialize(model_path,instance,transformation=Geom::Transformation.new)
@trans=transformation
@model_path=model_path
@is_group=false
@is_component=false
@instance=instance
@cost=0.0
@cost_source=nil
@parent_cost=nil
@mesh_bounds=Geom::BoundingBox.new
if Classify.profile_member?(instance)
	@profile_member=ProfileMember.new(instance)
else
	@profile_member=nil
	case instance
	when Sketchup::ComponentInstance
		@is_component=true
	when Sketchup::Group
		@is_group=true
	end
end
if !@profile_member
	get_mesh_bounds()
end

end
##
def contains_faces?
ents=@instance.definition.entities
face=ents.find {|e| e.is_a?(Sketchup::Face)}
if face
	return true
else
	return false
end
end

##
def reset_cost()
self.definition.delete_attribute(@@dict,'cost')
end
##
def set_cost(cost)
self.definition.set_attribute(@@dict,'cost',cost.to_f)
end
##
def local_transform
return @trans
end
##
def profile_member?
return true if @profile_member
end
##
def id
return @instance.entityID
end
##
def layer
model_path=@model_path.dup
model_path<<@instance
layers=Sketchup.active_model.layers
default_layer=layers[0]
model_path.each {|inst|
	default_layer=inst.layer if inst.layer!=layers[0]
}
layer=@instance.layer
if layer==layers[0]
	layer=default_layer
end
return layer
end
##
def visible?
model_path=@model_path.dup
model_path<<@instance
visible=true
model_path.each {|inst|
	visible=false if !inst.visible?
	visible=false if !inst.layer.visible?
}
return visible
end
##
def container
mp=@model_path.collect {|e| e.definition.name}
mp=mp.join("/")
return mp
end
##
def definition
return @instance.definition
end
##
def instance
return @instance
end
def profile_member
return @profile_member
end
##
def defn_name
return self.definition.name
end
##
def name
return @instance.name
end
##
def material
return @instance.material
end
##
def material_name
mat=self.material
if mat
	name=mat.name
else
	name="Default"
end
return name
end
##
def profile_name
if @profile_member
	name=@profile_member.profile.name
else
	name="-"
end
return name
end
#
def volume
vol=0.0
if @profile_member
	vol=@profile_member.volume
elsif @instance.manifold?
	vol=@instance.volume
else
	vol=get_mesh_volume()
end
return vol.abs
end
##
def get_mesh_bounds()
ents=@instance.definition.entities
faces=ents.grep(Sketchup::Face)
faces.each {|f| @mesh_bounds.add(f.bounds)}
end
##
def get_mesh_volume()
ents=@instance.definition.entities
edges=ents.grep(Sketchup::Edge)
edges=edges.find_all {|edge| edge.faces.length==1}
lengths=edges.collect {|edge| edge.length.to_f}
max_length=lengths.max
if max_length
	index=lengths.index(max_length)
	longest_edge=edges[index]
	p0=longest_edge.start.position.transform(@trans)
else
	p0=@mesh_bounds.center
end
faces=ents.grep(Sketchup::Face)
vol=0.0
faces.each {|f|
	mesh=f.mesh.transform!(@trans)
	polys=mesh.polygons
	polys.each {|poly|
		p1,p2,p3=mesh.point_at(poly[0]),mesh.point_at(poly[1]),mesh.point_at(poly[2])
		v0=p0.vector_to(p1)
		v1=p0.vector_to(p2)
		v2=p0.vector_to(p3)
		vol+=(v0.dot(v1.cross(v2)))/6.0
	}
}
bounds_vol=@mesh_bounds.width*@mesh_bounds.height*@mesh_bounds.depth
return [bounds_vol,vol].min
end
##
def area
length=self.length
vol=self.volume
bounds=@mesh_bounds
thickness=[bounds.width,bounds.height,bounds.depth].min
if @profile_member
	profile=@profile_member.profile
	width=profile.width
	height=profile.height
	if @profile_member.profile.is_1d?
		area=length*[width,height].max
	else
		area=vol/([length,width,height].min)
	end
elsif @instance.manifold?
	area=volume/thickness
else
	area=(bounds.width*bounds.height*bounds.depth)/thickness
end
return area
	
end
##
def height
height=0.0
if @profile_member
	height=@profile_member.profile.height
else
	bounds=@mesh_bounds
	height=[bounds.width,bounds.height,bounds.depth].min
end
return height
end
##
def width
width=0.0
if @profile_member
	width=@profile_member.profile.width
else
	bounds=@mesh_bounds
	height=[bounds.width,bounds.height,bounds.depth].min
	length=[bounds.width,bounds.height,bounds.depth].max
	width=[bounds.width,bounds.height,bounds.depth]
	width.delete(height)
	width.delete(length)
	width=width[0]
end
return width
end
##
def length
length=0.0
if @profile_member
	length=@profile_member.length(@trans)
else
	bounds=@mesh_bounds
	length=[bounds.width,bounds.height,bounds.depth].max
end
return length
end
##
def num_segments
if @profile_member
	num=@profile_member.num_segments
else
	num="-"
end
return num
end
##
def surface_area
area=0.0
if @profile_member
	area=@profile_member.surface_area
else
	ents=@instance.definition.entities
	faces=ents.grep(Sketchup::Face)
	faces.each {|f| area+=f.area}
end
return area
end
##
def weight
model_path=@model_path.dup
model_path<<@instance
layers=Sketchup.active_model.layers
default_layer=layers[0]
model_path.each {|inst|
	default_layer=inst.layer if inst.layer!=layers[0]
}
layer=@instance.layer
if layer==layers[0]
	layer=default_layer
end
layer_weight=LayerProperties.weight(layer).to_f
layer_weight_unit=LayerProperties.weight_unit(layer)
weight_lbs=0.0
case layer_weight_unit
	when "kg/m3","lb/ft3"
		weight_lbs=self.volume*Units.lbs_per_cubic_inch(layer_weight,layer_weight_unit)
	when "kg/m2","lb/ft2"
		weight_lbs=self.area*Units.psi(layer_weight,layer_weight_unit)
	when "kg/m","lb/ft"
		weight_lbs=self.length*Units.lbs_per_inch(layer_weight,layer_weight_unit)
end
return weight_lbs
end #func
##
def get_cost()
cost=get_parent_cost()
#p "parent cost is #{cost}"
return 0.0 if cost #if cost is inherited from a parent group or component, we don't want to add the cost of this entity
cost=get_instance_cost()
#p "instance cost is #{cost}"
return cost if cost
cost=get_definition_cost()
#p "definition cost is #{cost}"
return cost if cost
cost=get_layer_cost()
cost+=get_material_cost() #if cost
return cost
end #function
#
def get_instance_cost()
cost=@instance.get_attribute(@@dict,'cost',nil)
@cost_source="cost assigned to selected instance" if cost
return cost
end
##
def get_definition_cost()
defn=@instance.definition
cost=defn.get_attribute(@@dict,'cost',nil)
@cost_source="Cost assigned by user to selected Component" if cost
return cost
end
def get_layer_cost()
cost=0.0
layer=self.layer
layer_cost=LayerProperties.cost(layer).to_f
layer_cost_unit=LayerProperties.cost_unit(layer)
layer_weight=LayerProperties.weight(layer).to_f
layer_weight_unit=LayerProperties.weight_unit(layer)
unless self.contains_faces?
	@cost_source=nil
	return cost
end
if layer_cost==0.0
	@cost_source=nil
	return cost
end
case layer_cost_unit
	when "kg","lb","ton","tonne"
		case layer_weight_unit
			when "kg/m3","lb/ft3"
				vol=self.volume
				weight_lbs=vol*Units.lbs_per_cubic_inch(layer_weight,layer_weight_unit)
				cost_per_lb=Units.per_lb(layer_cost,layer_cost_unit)
				cost=cost_per_lb*weight_lbs
				@cost_source="#{Format.volume(vol)+Format.volume_unit(false)} * #{layer_weight}#{layer_weight_unit} * $#{layer_cost}/#{layer_cost_unit}"
			when "kg/m2","lb/ft2"
				area=self.area
				weight_lbs=area*Units.psi(layer_weight,layer_weight_unit)
				cost_per_lb=Units.per_lb(layer_cost,layer_cost_unit)
				cost=cost_per_lb*weight_lbs
				@cost_source="#{Format.area(area)+Format.area_unit(false)} * #{layer_weight}#{layer_weight_unit} * $#{layer_cost}/#{layer_cost_unit}"
			when "kg/m","lb/ft"
				len=self.length
				weight_lbs=len*Units.lbs_per_inch(layer_weight,layer_weight_unit)
				cost_per_lb=Units.per_lb(layer_cost,layer_cost_unit)
				cost=cost_per_lb*weight_lbs
				@cost_source="#{Format.length(len)+Format.length_unit} * #{layer_weight}#{layer_weight_unit} * $#{layer_cost}/#{layer_cost_unit}"
		end
	when "m2","ft2"
		area=self.area
		cost_per_square_in=Units.per_sq_inch(layer_cost,layer_cost_unit)
		cost=area*cost_per_square_in
		@cost_source="#{Format.area(area)+Format.area_unit(false)} * $#{layer_cost}/#{layer_cost_unit}"
	when "m3","ft3","yd3"
		vol=self.volume
		cost_per_cubic_inch=Units.per_cubic_inch(layer_cost,layer_cost_unit)
		cost=vol*cost_per_cubic_inch
		@cost_source="#{Format.volume(vol)+Format.volume_unit(false)} * $#{layer_cost}/#{layer_cost_unit}"
	when "m","ft"
		len=self.length
		cost_per_inch=Units.per_inch(layer_cost,layer_cost_unit)
		cost=len*cost_per_inch
		@cost_source="#{Format.length(len)+Format.length_unit} * $#{layer_cost}/#{layer_cost_unit}"
end #case
return cost
end
##
def get_material_cost()
default_material=nil
model_path=@model_path.dup
model_path<<@instance
model_path.each {|inst|
	default_material=inst.material if inst.material!=nil
}
	
defn=@instance.definition
ents=defn.entities
all_faces=ents.grep(Sketchup::Face)
face_mat={}
all_faces.each {|f|
	mat=f.material
	face_mat[mat] ||= []
	face_mat[mat].push(f)
}
finish_cost=0.0
face_mat.each {|mat,faces|
	area_sq_inch=0.0
	next if ((mat==nil) and (default_material==nil))
	mat=default_material if mat==nil
	mat_cost=MaterialProperties.cost(mat).to_f
	mat_cost_unit=MaterialProperties.cost_unit(mat)
	next if mat_cost==0.0
	
	case mat_cost_unit
		when "m2","ft2"
			faces.each {|f| area_sq_inch+=f.area}
			cost_per_sq_inch=Units.per_sq_inch(mat_cost,mat_cost_unit)
			cost=area_sq_inch*cost_per_sq_inch
			finish_cost+=cost
			@cost_source="" if not @cost_source
			@cost_source+="+" if @cost_source!=""
			@cost_source+="#{Format.area(area_sq_inch)+Format.area_unit(false)} * $#{mat_cost}/#{mat_cost_unit}"
	end #case
}
return finish_cost
end
##
def get_cost_source()
return @cost_source if @cost_source
end
private
##
def get_parent_cost()
@parent_cost=nil
@model_path.reverse.each {|inst|
	cost=inst.get_attribute(@@dict,'cost',nil)
	@cost_source="included in parent instance #{inst.name}" if cost
	unless cost
		defn=inst.definition
		cost=defn.get_attribute(@@dict,'cost',nil)
		@cost_source="included in parent definition #{defn.name}" if cost
	end
	@parent_cost=cost if cost
}
return @parent_cost
end
end #class Qmember
##
class QMemberList
public
def initialize()
@qmember_list=[]
end
##
def add(path,inst,trans)
q=QMember.new(path,inst,trans)
@qmember_list.push(q) if q.visible?
end
##
def clear
@qmember_list=[]
end
##
def each(&block)
@qmember_list.each(&block)
end
##
def [](index)
return @qmember_list[index]
end
##
def length
return @qmember_list.length
end
##
def count_profile_members()
pm=@qmember_list.find_all {|m| m.profile_member?}
return pm.length
end
#
def count_non_profile_members()
m=@qmember_list.find_all {|m| !(m.profile_member?)}
return m.length
end
##
def get_qmembers(ents)
@path=Sketchup.active_model.active_path
@path=[] unless @path
find_qmembers(ents,Geom::Transformation.new)
end
#
def find_qmembers(ents,trans)
ents.each {|e|
	if e.class==Sketchup::ComponentInstance
		self.add(@path.dup,e,trans*e.transformation)
		@path.push(e)
		find_qmembers(e.definition.entities,trans*e.transformation)
		@path.pop()
	elsif e.class==Sketchup::Group
		self.add(@path.dup,e,trans*e.transformation)
		@path.push(e)
		find_qmembers(e.entities,trans*e.transformation)
		@path.pop()
	end
}
			
end

end  #end class qmemberlist

class ProfileBuilderMTO  < Sketchup::SelectionObserver
include WebDialogHelper
private_class_method :new
@@profile_builder_mto=nil
@@dict="PBMTO"
def ProfileBuilderMTO.create
	@@profile_builder_mto=new unless @@profile_builder_mto
	return @@profile_builder_mto
end
def ProfileBuilderMTO.close()
if @@profile_builder_mto
	@@profile_builder_mto.close()
	@@profile_builder_mto=nil
end
end
def initialize
@qmember_list=QMemberList.new
@total_length=Hash.new(0.0)
@total_volume=Hash.new(0.0)
@total_members=Hash.new(0)
@total_segments=Hash.new(0)
@on_mac = (Object::RUBY_PLATFORM =~ /mswin/i) ? false : ((Object::RUBY_PLATFORM =~ /darwin/i) ? true : :other)
@report_dlg=nil
if ON_MAC
    @dlg=UI::WebDialog.new("PB Quantifier", false, "PB2Quantifier",350,280,250,250,true) #width,height,left,top
else
    @dlg=UI::WebDialog.new("PB Quantifier", false, "PB2Quantifier",350,340,250,250,true) #width,height,left,top
end
@dlg.set_file(File.join(PATH,"pb_mto.html"))
color = @dlg.get_default_dialog_color 
@dlg.set_background_color(color)
@dlg.set_on_close {remove_observers()}
@dlg.add_action_callback("ValueChanged") {|d,p|
	pair=p.split("|")
	id=pair[0]           
	value=pair[1]
	case id
	when "costOverride"
		value=@dlg.get_element_value("totalCost").to_f
		sel=Sketchup.active_model.selection
		inst=sel.first
		if inst.is_a?(Sketchup::Group) or inst.is_a?(Sketchup::ComponentInstance)
			qmember=QMember.new(Sketchup.active_model.active_path,inst)
			qmember.set_cost(value)
		end
		UI.messagebox("Cost of #{Format.cost(value)} assigned to selected object.")
		self.refresh_dialog(sel)
	when "costOverrideCancel"
		sel=Sketchup.active_model.selection
		inst=sel.first
		if inst.is_a?(Sketchup::Group) or inst.is_a?(Sketchup::ComponentInstance)
			qmember=QMember.new(Sketchup.active_model.active_path,inst)
			qmember.reset_cost()
		end
		self.refresh_dialog(sel)
	when "setLayers"
		LayerProperties.new.show()
	when "setMaterials"
		MaterialProperties.new.show()
	when "exportHTML"
		report_path=self.export_html()
		view=Sketchup.active_model.active_view
		width=view.vpwidth
		height=view.vpheight
		@report_dlg=UI::WebDialog.new("Profile Builder Report", false, nil,width,height,0,40,true) #width,height,left,top
		@report_dlg.set_file(report_path)
		@report_dlg.show {}
	when "exportCSV"
		export_csv_report()
	when "showMissingData"
		Sketchup.active_model.select_tool(ShowMissingCostDataTool.new)
	end
	
}
 
end
##
def show
Sketchup.active_model.select_tool(nil)
sel=Sketchup.active_model.selection
if ON_MAC == true
	@dlg.show_modal{refresh_dialog(sel)} unless @dlg.visible?
else
	@dlg.show{refresh_dialog(sel)} unless @dlg.visible?
end
sel.add_observer(self)
end
##
def close
remove_observers()
@dlg.close()
end
##
def validate_dialog()
sel=Sketchup.active_model.selection
#num_selected=@qmember_list.length
groups=sel.grep(Sketchup::Group)
instances=sel.grep(Sketchup::ComponentInstance)
num_selected=groups.length+instances.length
case num_selected
	when 0
		cmd="$('#costOverride').attr('disabled','disabled');"
	when 1
		qmember=QMember.new(Sketchup.active_model.active_path,@qmember_list[0])
		cmd="$('#costOverride').removeAttr('disabled');"
	else
		cmd="$('#costOverride').attr('disabled','disabled');"
end
@dlg.execute_script(cmd)
if num_selected==1
	if qmember.get_definition_cost()
		cmd="$('#costOverrideCancel').removeAttr('disabled');"
	else
		cmd="$('#costOverrideCancel').attr('disabled','disabled');"
	end
else
	cmd="$('#costOverrideCancel').attr('disabled','disabled');"
end
@dlg.execute_script(cmd)
end
##
def export_csv_report()
ext=".csv"
		
path=UI.savepanel("Export CSV Report", "","")
return if not path
f_name=File.basename(path,".*")
f_name+=ext
path=File.dirname(path)+"\\"+f_name
#path=File.join(File.dirname(__FILE__),"report.csv")
begin
	out=File.new(path,"w")
	write_report_csv(out)
	out.close()
	UI.openURL(path)
rescue
	UI.messagebox("Unable to export report.")
	raise
ensure
	out.close() unless out.closed?
end
end
##
def set_table(html)
cmd="$('#reportContainer').html(html);"
@report_dlg.execute_script(cmd)
end
def create_report()
cmd="createTable();"
@report_dlg.execute_script(cmd)
end
def remove_observers()
Sketchup.active_model.selection.remove_observer(self)
end
####
def onSelectionBulkChange(selection)
   refresh_dialog(selection)
end
def onSelectionCleared(selection)
    refresh_dialog(selection)
end
##
def refresh_dialog(selection)
reset()
@qmember_list.get_qmembers(selection)
volume=total_volume()
area=total_area(selection)
weight=total_weight()
length=total_length()
cost=total_cost()
cost+=get_finish_cost(selection)
cost_source="No Cost Data"
if selection.length==1
	cost_source=get_cost_source(selection[0])
	cost_source="No Cost Data" if cost_source==nil
else
	cost_source=""
end
setLabels()
setValue("numProfileMembers",@qmember_list.count_profile_members())
setValue("numInstances",@qmember_list.count_non_profile_members())
setValue("totalLength",Format.length(length))
setValue("totalVolume",Format.volume(volume))
setValue("totalWeight",Format.weight(weight))
setValue("totalSurfaceArea",Format.area(area))
setValue("totalCost",Format.cost(cost))
setValue("costSource",cost_source)
validate_dialog()
end
##
def setLabels
cmd="$('#lengthLabel').html('Length (#{Format.length_unit})');"
@dlg.execute_script(cmd)
cmd="$('#volumeLabel').html('Volume (#{Format.volume_unit})');"
@dlg.execute_script(cmd)
cmd="$('#areaLabel').html('Surface Area (#{Format.area_unit})');"
@dlg.execute_script(cmd)
cmd="$('#weightLabel').html('Weight (#{Format.weight_unit})');"
@dlg.execute_script(cmd)
end
##
def reset()
@qmember_list.clear
@total_length=Hash.new(0.0)
@total_volume=Hash.new(0.0)
@total_members=Hash.new(0)
@total_segments=Hash.new(0)
@cost_source=nil
end
##
def get_cost_source(ent)
path=Sketchup.active_model.active_path
path=[] unless path
cost_source=nil
case ent
	when Sketchup::ComponentInstance, Sketchup::Group
		qmember=QMember.new(path,ent)
		cost=qmember.get_cost()
		cost_source=qmember.get_cost_source()
	when Sketchup::Face
		cost_source=@cost_source
	else
		cost_source=""
end
return cost_source
end
##
def get_finish_cost(ents)
all_faces=ents.grep(Sketchup::Face)
face_mat={}
all_faces.each {|f| mat=f.material
	face_mat[mat] ||= []
	face_mat[mat].push(f)
}
finish_cost=0.0
area_sq_inch=0.0
face_mat.each {|mat,faces|
	next if mat==nil
	mat_cost=MaterialProperties.cost(mat).to_f
	mat_cost_unit=MaterialProperties.cost_unit(mat)
	next if mat_cost==0.0
	case mat_cost_unit
		when "m2","ft2"
			faces.each {|f| area_sq_inch+=f.area}
			cost_per_sq_inch=Units.per_sq_inch(mat_cost,mat_cost_unit)
			cost=area_sq_inch*cost_per_sq_inch
			finish_cost+=cost
	end
	if all_faces.length==1
		@cost_source="" if not @cost_source
		@cost_source+="#{Format.area(area_sq_inch)+Format.area_unit(false)} * $#{mat_cost}/#{mat_cost_unit}" 
	end
}
return finish_cost
end
########################
def get_totals()
@qmembers.each {|m|
	path=m[0]
	pm=m[1]
	trans=m[2]
	p_name=pm.profile.name
	@total_length[p_name]+=pm.length(trans)
	@total_volume[p_name]+=pm.volume(trans)
	@total_members[p_name]+=1
	@total_segments[p_name]+=pm.num_segments
}
end
###
def total_volume()
total_volume=0.0
@qmember_list.each {|member|
	total_volume+=member.volume
}
return total_volume
end
##
def total_length()
total_length=0.0
@qmember_list.each {|member|
	total_length+=member.length
}
return total_length
end
##
def total_area(selection)
total_area=0.0
@qmember_list.each {|member|
	total_area+=member.surface_area
}
faces=selection.grep(Sketchup::Face)
faces.each {|f| total_area+=f.area}
return total_area
end

##
def total_weight()
total_weight=0.0
@qmember_list.each {|member|
	total_weight+=member.weight
}
return total_weight
end
##
def total_cost()
total_cost=0.0
@qmember_list.each {|member|
	cost=member.get_cost()
	if cost
		total_cost+=cost
	end
}

return total_cost
end
##
def html_td(text)
string="<td>#{text}</td>"
return string
end
##
def html_th(text)
string="<th>#{text}</th>"
return string
end
def write_report_header_html(out)
root=File.dirname(__FILE__)
data_tables_css="DataTables-1.10.4/media/css/jquery.dataTables.css"
data_tables_jquery="DataTables-1.10.4/media/js/jquery.js"
data_tables_js="DataTables-1.10.4/media/js/jquery.dataTables.js"
select_table_js="DataTables-1.10.4/media/js/selectTable.js"
out.puts "<head>"
out.puts "<meta charset='UTF-8'>"
out.puts "<link rel='stylesheet' type='text/css' href='#{data_tables_css}'>"
out.puts "<script type='text/javascript' charset='utf8' src='#{data_tables_jquery}'></script>"
out.puts "<script type='text/javascript' charset='utf8' src='#{data_tables_js}'></script>"
out.puts "<script type='text/javascript' charset='utf8' src='#{select_table_js}'></script>"
out.puts "</head>"
end
##
def write_report_table_initializer(out)
out.puts "<script type='text/javascript'>
$(document).ready(function() {
    $('#report').dataTable( {
     \"lengthMenu\": [[20, 50, 100,-1], [20, 50,100, \"All\"]]
    } );
} );
</script>"
end
##
def report_row_data(m)
[m.layer.name,m.container,m.defn_name,m.name,m.material_name,m.profile_name,Format.small_length(m.width),Format.small_length(m.height),Format.length(m.length),Format.volume(m.volume),Format.area(m.area),m.num_segments,Format.area(m.surface_area),Format.weight(m.weight),Format.cost(m.get_cost()),Format.cost(m.get_material_cost()),Format.cost(m.get_cost())]
end
##
def report_headers(html=true)
["Layer","Container","Component Name","Object Name","Object Material","Profile","Width (#{Format.small_length_unit})","Height (#{Format.small_length_unit})","Length (#{Format.length_unit})","Volume (#{Format.volume_unit(html)})","Area (#{Format.area_unit(html)})","Segments","Surface Area (#{Format.area_unit(html)})","Weight (#{Format.weight_unit})","Object Cost","Surface Material Cost","Total Cost"]
end
##
def write_report_table_html(out)
headers=report_headers()
out.puts "<table id='report' cellspacing='0' width='100%' class='display' style='text-align:center'>"
out.puts "<thead>"
out.puts "<tr>"
headers.each {|text| out.puts(html_th(text))}
out.puts "</tr>"
out.puts "</thead>"
out.puts "<tbody>"
@qmember_list.each {|m|
	next unless m.contains_faces?
	out.puts "<tr>"
	row_data=report_row_data(m)
	row_data.each {|data| out.puts(html_td(data))}
	out.puts "</tr>"
}
out.puts "</tbody>"
out.puts "</table>"
end
##
def write_select_button(out)
out.puts "<input type='button' value='Select Table' style='float:right;padding:4px;margin:4px'
   onclick=\"selectElementContents( document.getElementById('report') );\">"
end
###################
def export_html()
report_path=File.join(File.dirname(__FILE__),"report.html")
out=File.new(report_path,"w")
begin
	out.puts "<!DOCTYPE html>"
	out.puts "<html>"
	write_report_header_html(out)
	out.puts "<body style='font-family:arial;font-size:0.8em;overflow:scroll'>"
	write_report_table_initializer(out)
	write_select_button(out)
	write_report_table_html(out)
	out.puts "</body>"
	out.puts "</html>"
rescue
	raise
ensure
	out.close()
end
return report_path
end
##
def write_report_csv(out)
html=false
headers=report_headers(html)
out.puts(headers.join(","))
@qmember_list.each {|m|
	next unless m.contains_faces?
	row_data=report_row_data(m)
	out.puts(row_data.join(","))
}
end
end  #class end
#############################
class Units
#
def self.lbs_per_cubic_inch(weight,unit)
factor=1.0
case unit
	when "kg/m3"
		factor=0.000036127292153
	when "lb/ft3"
		factor=0.0005787037037
end
return weight*factor
end
##
def self.per_lb(num,unit)
factor=1.0
case unit
	when "kg"
		factor=2.2046226218
	when "lb"
		factor=1.0
	when "ton"
		factor=2000.0
	when "tonne"
		factor=2204.6226218
end
return num*(1.0/factor)
end
##
def self.psi(weight,unit)
factor=1.0
case unit
	when "kg/m2"
		factor=0.0014223343334
	when "lb/ft2"
		factor=0.0069444444445
end
return weight*factor
end

##
def self.lbs_per_inch(weight,unit)
factor=1.0
case unit
	when "kg/m"
		factor=0.055880000000
	when "lb/ft"
		factor=0.083333333333
end
return weight*factor
end
##
def self.per_cubic_inch(num,unit)
factor=1.0
case unit
	when "m3"
		factor=61023.744095
	when "ft3"
		factor=1728.0
	when "yd3"
		factor=46656.0
end
return num*(1.0/factor)
end
##
def self.per_sq_inch(num,unit)
factor=1.0
case unit
	when "m2"
		factor=1550.0031
	when "ft2"
		factor=144.0
end
return num*(1.0/factor)
end
##
def self.per_inch(num,unit)
factor=1.0
case unit
	when "m"
		factor=39.37007874
	when "ft"
		factor=12.0
end
return num*(1.0/factor)
end

end #class units
##
class MaterialProperties
include WebDialogHelper
@@dict="PBMTO"
def initialize
@dlg=UI::WebDialog.new("Material Data", false, "PB2MatCost",300,185,250,250,true) #width,height,left,top
@dlg.set_file(File.join(PATH,"material_cost.html"))
color = @dlg.get_default_dialog_color 
@dlg.set_background_color(color)
model = Sketchup.active_model
@materials = model.materials
if @materials.length==0
	UI.messagebox("Please create at least one material.")
	raise "No Materials in Model."
end
@dlg.add_action_callback("ValueChanged") {|d,p|
	pair=p.split("|")
	id=pair[0]
	value=pair[1]
	material=self.get_current_material()
	case id
		when "material_cost"
			value=value.to_f
			material.set_attribute(@@dict, 'cost', value)
		when "materialCostUnit"
			material.set_attribute(@@dict, 'cost_unit', value)
		when "materialName"
			if @materials[value]
				self.current_material=@materials[value]
			end
			refresh_dialog()
		when "closeButton"
			@dlg.close()
	end
}
end
def show
material=self.get_current_material()
@dlg.show_modal{refreshMaterialsList("materialName",false);setValue("materialName",material.name);refresh_dialog()}
end
##
def refresh_dialog()
material=self.get_current_material()
setValue("material_cost",MaterialProperties.cost(material))
setValue("materialCostUnit",MaterialProperties.cost_unit(material))
end
#
def self.cost(mat)
return 0.0 if mat==nil
Format.cost(mat.get_attribute(@@dict,'cost',0.0))
end
def self.cost_unit(mat)
default_cost_unit=Format.area_unit(false)
return "" if mat==nil
mat.get_attribute(@@dict,'cost_unit',default_cost_unit)
end
##
def get_current_material()
mat=nil
model=Sketchup.active_model
mats=model.materials
mat_name=model.get_attribute(@@dict,'current_material')
if mat_name and mat_name!=""
	mat=mats[mat_name]
end
if mat
	return mat
elsif mats.length>0
	return mats[0]
else
	UI.messagebox("Please create at least one material.")
	raise "No Materials in Model."
end
end
##
def current_material=(mat)
Sketchup.active_model.set_attribute(@@dict,'current_material',mat.name)
end
end #class material properties
#############################################
class LayerProperties
include WebDialogHelper
@@dict="PBMTO"

def initialize
@dlg=UI::WebDialog.new("Layer Data", false, "PB2LayerData",300,185,250,250,true) #width,height,left,top
@dlg.set_file(File.join(PATH,"layer_cost.html"))
color = @dlg.get_default_dialog_color 
@dlg.set_background_color(color)
@dlg.add_action_callback("ValueChanged") {|d,p|
	pair=p.split("|")
	id=pair[0]
	value=pair[1]
	layer=self.get_current_layer()
	case id
		when "layerCost"
			value=value.to_f
			layer.set_attribute(@@dict, 'cost', value)
		when "layerCostUnit"
			layer.set_attribute(@@dict, 'cost_unit', value)
		when "layerWeight"
			value=value.to_f
			layer.set_attribute(@@dict, 'weight', value)
		when "laterWeightUnit"
			layer.set_attribute(@@dict, 'weight_unit', value)
		when "layerName"
			model = Sketchup.active_model
			layers=model.layers
			if layers[value]
				self.current_layer=layers[value]
			end
			refresh_dialog()			
		when "closeButton"
			@dlg.close()
	end
}
end
##
def show
layer=get_current_layer()
@dlg.show_modal{refreshLayersList("layerName",true);setValue("layerName",layer.name);refresh_dialog()}
end
##
def refresh_dialog()
layer=get_current_layer()
setValue("layerCost",LayerProperties.cost(layer))
setValue("layerCostUnit",LayerProperties.cost_unit(layer))
setValue("layerWeight",LayerProperties.weight(layer))
setValue("laterWeightUnit",LayerProperties.weight_unit(layer))
end
###
def self.weight(layer)
return 0.0 if layer==nil
layer.get_attribute(@@dict,'weight',0.0)
end
#
def self.cost(layer)
return 0.0 if layer==nil
Format.cost(layer.get_attribute(@@dict,'cost',0.0))
end
#
def self.weight_unit(layer)
default_weight_unit=Format.default_density_unit(false)
return "" if layer==nil
layer.get_attribute(@@dict,'weight_unit',default_weight_unit)
end
def self.cost_unit(layer)
default_cost_unit=Format.default_cost_unit
return "" if layer==nil
layer.get_attribute(@@dict,'cost_unit',default_cost_unit)
end
##
def get_current_layer()
layer=nil
model=Sketchup.active_model
layers=model.layers
layer_name=model.get_attribute(@@dict,'current_layer')
if layer_name and layer_name!=""
	layer=layers[layer_name]
end
if layer
	return layer
elsif layers.length>0
	return layers[0]
else
	UI.messagebox("Please create at least one layer.")
	raise "No Layers in Model."
end
end
##
def current_layer=(layer)
Sketchup.active_model.set_attribute(@@dict,'current_layer',layer.name)
end
end #class LayerProperties
##
class ShowMissingCostDataTool
include Geometry
def initialize
@model=Sketchup.active_model
@ents=@model.active_entities
@missing_cost_data=[]
@has_cost_data=[]
@qmember_list=QMemberList.new
@previewer=ComponentPreviewer.new
analyze_model()	
end
##
def analyze_model()
Sketchup.active_model.selection.clear
@qmember_list.get_qmembers(@ents)
@qmember_list.each {|m|
	cost=m.get_cost()
	cost_source=m.get_cost_source()
	if cost_source==nil
		@missing_cost_data.push(m)
	else
		@has_cost_data.push(m)
	end
}
end
##
def onMouseMove(flags, x, y, view)
view.invalidate
end
##
def onLButtonDown(flags,x,y,view)
sel=Sketchup.active_model.selection
sel.clear
best=get_inst_below_cursor(view,x,y)
if best
	sel.add(best)
	ProfileBuilderMTO.create.refresh_dialog(sel)	
end
end
##
def activate
Sketchup.active_model.active_view.invalidate
end
##
def draw(view)

view.drawing_color="red"
view.line_width=3
@missing_cost_data.each {|m|
	defn=m.definition
	@previewer.definition=defn
	@previewer.transformation=m.local_transform
	@previewer.draw_bounds()
}
view.drawing_color="green"
view.line_width=3
@has_cost_data.each {|m|
	defn=m.definition
	@previewer.definition=defn
	@previewer.transformation=m.local_transform
	@previewer.draw_bounds()
}
end
end  #class
end #module profile builder
end #module DM
