require 'sketchup.rb'
module DM
module ProfileBuilder2
class ProfileMember
include Geometry

#defaults
@@dict="ProfileBuilder"
@@pp=5
@@rot=0.0
@@mirror=false
@@chain=[]
@@x_offset=0.0
@@y_offset=0.0
@@smooth_angle=30.0
@@material_name=nil
@@layer_name=Sketchup.active_model.layers[0].name
#adds a new profile member group to the model
def self.add(ents,chain,trans=nil)
group=ents.add_group
pm=ProfileMember.new(group)
trans=default_transformation(chain) if trans==nil
pm.transform!(trans)
#transformed_chain=chain.collect {|p| p.transform(trans.inverse).to_a}
pm.set_chain(chain,trans)
return pm
end
#
def self.default_transformation(chain)
if chain.length>2
	trans=Geom::Transformation.new
else
	pm_origin=chain.first
	pm_zaxis=chain[1]-chain[0]
	trans=Geom::Transformation.new(pm_origin,pm_zaxis)
	trans=Geom::Transformation.new(trans.xaxis.reverse,trans.zaxis,trans.yaxis,pm_origin)
end
return trans
end
private_class_method :default_transformation
##
def initialize(inst)
@group=inst
@su_defn=@group.definition if inst.class==Sketchup::Group
@su_defn=@group.definition if inst.class==Sketchup::ComponentInstance
end
public
##
def reverse()
path=self.chain.path
new_path=path.reverse
self.set_chain(new_path)
self.draw()
end
##
def name=(value)
@group.name=value
end
##
def name
return @group.name
end
##
def hide
@group.hidden=true
end
##
def show
@group.hidden=false
end
###
def definition
return @su_defn
end
###
def <=>(other)
return self.profile.name<=>other.profile.name
end
#applies all of the profile settings to the profile member
def set_from_profile!(profile)
self.name=profile.name
self.profile=profile
self.rotation=profile.rotation
self.mirror=profile.mirror
self.placement_point=profile.placement_point
self.x_offset=profile.x_offset
self.y_offset=profile.y_offset
self.smooth_angle=profile.smooth_angle
self.x_scale=profile.x_scale
self.y_scale=profile.y_scale
self.material_name=profile.material_name
self.layer=get_layer_by_name(profile.layer_name)
end
##
def revolved=(value)
@su_defn.set_attribute(@@dict,'revolved',value)
end
def revolved?
@su_defn.get_attribute(@@dict,'revolved',false)
end
##
def revolve_segments=(value)
@su_defn.set_attribute(@@dict,'revolve_segments',value.to_i)
end
##
def revolve_segments
@su_defn.get_attribute(@@dict,'revolve_segments',24)
end
############get the group object that contains the member
def instance
return @group
end
###########set the group that will contain the member
def instance=(g)
@group=g
end
###############set the profile of the member
def profile=(p)
@su_defn.set_attribute(@@dict,'profile',p.to_s)
end
#####get the entities object if this member
def entities
if @group.class==Sketchup::Group
	return @group.entities
else
	return @group.definition.entities
end
end
###############get the profile of the member
def profile
p_string=@su_defn.get_attribute(@@dict,'profile')||@group.get_attribute(@@dict,'profile')
if p_string
	return Profile.new(p_string)
else
	return nil
end
end
##
def material
return @group.material
end
##
def material_name=(name)
@su_defn.set_attribute(@@dict,'material_name',name)
end
##
def material_name
@su_defn.get_attribute(@@dict,'material_name')||@group.get_attribute(@@dict,'material_name')||@@material_name
end
##
def layer_name
if @group.layer
	name=@group.layer.name
else
	name=Sketchup.active_model.layers[0].name
end
return name
#@su_defn.get_attribute(@@dict,'layer_name')||@group.get_attribute(@@dict,'layer_name')||@@layer_name
end
##
def layer
@group.layer
end
##
def layer=(value)
unless value.is_a?(Sketchup::Layer) or value.is_a?(String)
	raise ArgumentError,"Layer or String class expected",caller
end
if value.is_a?(Sketchup::Layer)
	@group.layer=(value)
else	
	@group.layer=get_layer_by_name(value)
end
end
#################
def placement_point=(pp)  #pp must be an integer between 1 and 9
@su_defn.set_attribute(@@dict,'pp',pp)
end
##################
def placement_point
@su_defn.get_attribute(@@dict,'pp')||@group.get_attribute(@@dict,'pp')||@@pp
end
##################
def rotation=(rot)
@su_defn.set_attribute(@@dict,'rotation',rot)
end
##################
def rotation
@su_defn.get_attribute(@@dict,'rotation')||@group.get_attribute(@@dict,'rotation')||@@rot
end
##################
def x_offset=(value)
@su_defn.set_attribute(@@dict,'x_offset',value)
end
##################
def x_offset
@su_defn.get_attribute(@@dict,'x_offset')||@group.get_attribute(@@dict,'x_offset')||@@x_offset
end
##################
def y_offset=(value)
@su_defn.set_attribute(@@dict,'y_offset',value)
end
##################
def y_offset
@su_defn.get_attribute(@@dict,'y_offset')||@group.get_attribute(@@dict,'y_offset')||@@y_offset
end
##################
def x_scale=(value)
@su_defn.set_attribute(@@dict,'x_scale',value.to_f)
end
##################
def x_scale
@su_defn.get_attribute(@@dict,'x_scale',1.0)
end
##################
def y_scale=(value)
@su_defn.set_attribute(@@dict,'y_scale',value.to_f)
end
##################
def y_scale
@su_defn.get_attribute(@@dict,'y_scale',1.0)
end

##################
def smooth_angle=(value)
@su_defn.set_attribute(@@dict,'smooth_angle',value)
end
##################
def smooth_angle
@su_defn.get_attribute(@@dict,'smooth_angle')||@group.get_attribute(@@dict,'smooth_angle')||@@smooth_angle
end
##################
def mirror=(value)
@su_defn.set_attribute(@@dict,'mirror',value)
end
#################
def mirror
@su_defn.get_attribute(@@dict,'mirror')||@group.get_attribute(@@dict,'mirror')||@@mirror
end
########
def set_chain(chain,trans=nil) 
trans=Geom::Transformation.new unless trans
pts=chain.collect {|p| p.transform(trans.inverse).to_a}
@su_defn.set_attribute(@@dict,'chain',pts)
end
#################
def chain
pts=@su_defn.get_attribute(@@dict,'chain')||@group.get_attribute(@@dict,'chain')||@@chain
pts=pts.collect {|p| Geom::Point3d.new(p)}
return Chain.new(pts)
end
def closed_path?
path=self.chain
return path.first==path.last
end
#################get the length of the member in inches
def length(trans=Geom::Transformation.new)
len=0.0
pts=self.chain
return 0.0 if pts.length<2
pts=pts.collect {|p| Geom::Point3d.new(p[0],p[1],p[2]).transform(trans)}
(pts.length-1).times {|i|
	p1=pts[i]
	p2=pts[i+1]
	vec=p1.vector_to(p2)
	len+=vec.length
}
return len
end
##
def surface_area
area=0.0
faces=self.entities.grep(Sketchup::Face)
faces.each {|f| area+=f.area}
return area
end
################get the volume of the member in cubic inches
def volume(transformation=Geom::Transformation.new)
volume=@group.volume  
if volume<0.0
	area=self.profile.area(transformation)
	volume=area*self.length(transformation)
end
return volume
end
####################get the number of segments that make up the member
def num_segments
return self.chain.length-1
end
######################move the profile member to the given point and point the z axis the same direction as the given vector
def move!(pt,vec)   
trans=Geom::Transformation.new(pt)
@group.transformation=trans
end
def transform!(trans)
@group.transformation=trans
end
##
def extend_start(length,split_member=false)
#normal=get_normal_of_vertices(verts)
normal=start_extend_vector()
normal.length=length
extend_transform=Geom::Transformation.new(normal)
ents=self.entities
path=self.chain
unless length>0 and split_member
	if self.profile.is_1d?
		verts=self.get_cap_vertices().first
		curve_edge=verts[0].edges.find {|edge| edge.curve!=nil}
		curve=curve_edge.curve
		new_positions=verts.collect {|v| v.position.transform(extend_transform)}
		curve.move_vertices(new_positions)
	else	
		verts=self.get_cap_face(true).vertices
		ents.transform_entities(extend_transform,verts)
	end
	path=self.chain
	path[0]=path[0].offset(normal)
	self.set_chain(path)
end
if split_member
	self.split(path[0].offset(normal),path[0]) if length>0
	self.split(path[0].offset(normal.reverse),path[0]) if length<0
end
end
##
def extend_end(length,split_member=false)

#normal=get_normal_of_vertices(verts)
normal=end_extend_vector()
normal.length=length
extend_transform=Geom::Transformation.new(normal)
ents=self.entities
path=self.chain
unless length>0 and split_member
	if self.profile.is_1d?
		verts=self.get_cap_vertices().last
		curve_edge=verts[0].edges.find {|edge| edge.curve!=nil}
		curve=curve_edge.curve
		new_positions=verts.collect {|v| v.position.transform(extend_transform)}
		curve.move_vertices(new_positions)
	else
		verts=self.get_cap_face(false).vertices
		ents.transform_entities(extend_transform,verts)
	end
	path[-1]=path[-1].offset(normal)
	self.set_chain(path)
end
if split_member
	self.split(path[-1],path[-1].offset(normal)) if length>0
	self.split(path[-1],path[-1].offset(normal.reverse)) if length<0
end
end
##
def split(p1,p2)
p1=Geom::Point3d.new(p1)
p2=Geom::Point3d.new(p2)
new_chain=Chain.new
trans=self.transformation
new_chain.set_path([p1.transform(trans),p2.transform(trans)])
entities=Sketchup.active_model.active_entities
pm=ProfileMember.add(entities,new_chain.path)
pm.name=self.instance.name
copy_attributes_to(pm.instance)
pm.set_chain(new_chain,pm.transformation)
pm.draw()
end
##
def draw()
clear_existing_edges()
path_extruder=PathExtruder.new
path_extruder.path=self.chain.path
profile=self.profile
profile.set_from_profile_member(self)
path_extruder.profile=profile
path_extruder.material=get_material()
path_extruder.smooth_angle=self.smooth_angle
path_extruder.extrude_along_path(self.entities)
soften_edges()
clear_stray_edges()
end
##
def clear_stray_edges()
ents=self.entities
edges=ents.grep(Sketchup::Edge)
strays=edges.find_all {|edge| edge.faces.length==0}
ents.erase_entities(strays)
end
###
def draw_path()
clear_existing_edges()
chain=self.chain
ents=self.entities
edit_trans=Sketchup.active_model.edit_transform
edges=[]
chain.edges.each {|edge|
	p0=edge[0].transform(edit_trans)
	p1=edge[1].transform(edit_trans)
	edges.push(ents.add_line([p0,p1]))
}
return edges
end
#
def draw_path_group(trans)
pts=self.chain.path
group=Sketchup.active_model.active_entities.add_group
group_ents=group.entities
group.transform! trans
group_ents.add_edges(pts)
return group
end
##
def pmpi(trans)
@group.hidden=true
@group.set_attribute('PMPI','pmpi_hidden',true)
path=self.draw_path_group(trans)
path.set_attribute('PMPI','pmpi_path',true)
end
##
def soften_edges()
if self.smooth_angle>0.0
	mat=get_material()
	if mat.respond_to?(:texture) and mat.texture!=nil
		soften(smooth_angle,false)
	else
		soften(smooth_angle,true)
	end
end
end
##
def set_material()
@group.material=get_material()
end
#revolve a profile around an axis to create the profile member
def revolve()
clear_existing_edges()
ents=self.entities
profile=self.profile
profile.set_from_profile_member(self)
pm_revolver=ProfileRevolver.new(profile)
mat=get_material()
pm_revolver.material=mat
mesh,uv_hash=pm_revolver.get_revolve_mesh(self.revolve_segments)
if mat and mat.texture
	uv_hash.each {|poly_index,uvs|
		pts=mesh.polygon_points_at(poly_index)
		poly=mesh.polygon_at(poly_index)
		f=ents.add_face(pts)
		pt_array=[]
		poly.each {|index| pt_array.push(mesh.point_at(index));pt_array.push(uvs[index])}
		f.position_material(mat,pt_array,true)
	}
else
	ents.add_faces_from_mesh(mesh,Geom::PolygonMesh::AUTO_SOFTEN|Geom::PolygonMesh::SMOOTH_SOFT_EDGES,mat)
end
self.revolved=true
soften_edges()
end #revolve
##
def get_cap_face(start)
cap_face=nil
ents=self.entities
faces=ents.grep(Sketchup::Face)
profile=self.profile
num_points=profile.points.length
profile_area=profile.area
if start
	cap_normal=start_extend_vector()
	point=self.chain[0]
else
	cap_normal=end_extend_vector()
	point=self.chain[-1]
end
faces=faces.find_all {|f| f.outer_loop.vertices.length==num_points}
faces=faces.find_all {|f| f.normal.angle_between(cap_normal)<(Math::PI/2.0)} #allows tapered cap faces
faces=faces.find_all {|f| ((projected_face_area(f,cap_normal)-self.profile_area)).abs<0.001}
=begin
faces=faces.find_all {|f|
	cp_normal=cap_normal
	edges=f.edges
	adjacent_faces=[]
	edges.each {|edge|
		edge_faces=edge.faces
		edge_faces.delete(f)
		adjacent_faces.push(edge_faces[0])
	}
	perp_adjacent_faces=adjacent_faces.find_all {|af| af.normal.perpendicular?(cp_normal)}
	perp_adjacent_faces.length==adjacent_faces.length
}
=end
if faces.length>1
	cap_face=get_closest_face_to_point(faces,point)
else
	cap_face=faces[0]
end
		
return cap_face
end
##
def profile_area
profile=self.profile
t=transform_by_size(profile.width,profile.height,self.x_scale,self.y_scale)
area=profile.area(t)
return area
end
###################get the transformation of the profile member
def transformation
return self.instance.transformation
end
#####################
def xscale
    ((Geom::Vector3d.new 1,0,0).transform! self.transformation).length
end
def yscale
    ((Geom::Vector3d.new 0,1,0).transform! self.transformation).length
end
def zscale
    ((Geom::Vector3d.new 0,0,1).transform! self.transformation).length
end

####
def to_path!
group_ents=self.entities
edges=group_ents.find_all {|e| e.class==Sketchup::Edge}
group_ents.erase_entities(edges)   #remove all edges and faces inside the group
pts=self.chain
group_ents.add_edges(pts)
end
##
def trim_to(inst,cap_index)
inst=validate_for_trim(inst)
prepare_end_for_trim(inst,cap_index)
result=perform_boolean_trim(inst)
delete_waste(result)
end
##
def get_picked_cap_face_index(pickray)
chain=self.chain
trans=self.transformation
closest=chain.closest_point_on_path(pickray,trans)
dist_on_path=chain.distance_along_path(closest[0],trans)
length=self.length
if dist_on_path==nil
	#p "point is not on path"
	return 0
end
ratio=dist_on_path/length
if ratio<0.5
	index=0
else
	index=1
end
return index
end
##
def get_closest_path_point(pickray)
chain=self.chain
edges=chain.edges
trans=self.transformation
closest=chain.closest_point_on_path(pickray,trans)
pt=closest[0]
edge_index=closest[1]
edge=edges[edge_index]
p0=edge[0].transform(trans)
p1=edge[1].transform(trans)
edge_length=p0.distance(p1)
dist_to_pt=p0.distance(pt)
ratio=dist_to_pt/edge_length
if ratio<=0.5
	pt_index=chain.index(edge[0])
else
	pt_index=chain.index(edge[1])
end
return pt_index
end
##
def trim_to_plane(plane,cap_index)
@group=@group.make_unique if @group.kind_of?(Sketchup::Group)
@su_defn=@group.definition
cap_verts=self.get_cap_vertices[cap_index]
dist=get_max_distance_to_plane(cap_verts,plane,self.transformation)
if !dist
	raise TrimException, "不能修剪/延伸.轮廓放样的路径与所选的面不相交。",caller
end
perp_edges=get_perpindicular_edges(cap_verts)
perp_edges_lengths=perp_edges.collect {|pe| pe.length.to_f}
min_length=perp_edges_lengths.min
if  (dist<0.0) and ((dist.abs)>min_length)
	raise TrimException, "不能修剪/延伸.选择的面与选择的轮廓放样模型的端头处于相反方向，请选择另个端头。", caller
end
if cap_index==0
	self.extend_start(dist)
elsif cap_index==1
	self.extend_end(dist)
end
project_vertices_to_plane(cap_verts,plane)
end
##
def get_cap_vertices()
if self.profile.is_1d?
	caps=[get_cap_curve_vertices(true),get_cap_curve_vertices(false)]
else
	caps=get_cap_faces()
	caps[0]=caps[0].outer_loop.vertices if caps[0]
	caps[1]=caps[1].outer_loop.vertices if caps[1]
end
return caps
end
##
def get_cap_faces()
return [get_cap_face(true),get_cap_face(false)]
end
##
def get_perpendicular_cap_faces()
faces=self.get_cap_faces()
if faces[0]
	faces[0]=nil if !(faces[0].normal.samedirection?(start_extend_vector()))
end
if faces[1]
	faces[1]=nil if !(faces[1].normal.samedirection?(end_extend_vector()))
end
return faces
end
def end_extend_vector()
path=self.chain
return path[-2].vector_to(path[-1]).normalize
end
###
def start_extend_vector()
path=self.chain
return path[1].vector_to(path[0]).normalize
end
	
#################
private
	
	##for backward compatibility
	##from version 2.0, all attributes are stored at the definition level
	def copy_group_dict_to_defn_deprecated()
	
	dict=nil
	dicts=@group.attribute_dictionaries
	dict=dicts[@@dict] if dicts
	if dict
		#Sketchup.active_model.start_operation("Upgrade PM",true,true)
		dict.each {|key,val|
			if key and key!=""
				#p "copying attribute #{key} to defn"
				@su_defn.set_attribute(@@dict,key,val)
			end
		}
		@group.delete_attribute(@@dict)
	end
	end
	##
	def copy_attributes_to(dest_instance)
	dicts=@su_defn.attribute_dictionaries
	dest_defn=dest_instance.definition
	dest_instance.layer=self.layer
	if dicts
		dicts.each {|dict|
			dict.each_pair {|key,value| dest_defn.set_attribute(dict.name,key,value)}
		}
	end
		
	end
	#
	def clear_existing_edges()
	@group.make_unique if @group.kind_of?(Sketchup::Group) 
	group_ents=self.entities
	edges=group_ents.grep(Sketchup::Edge)
	text=group_ents.grep(Sketchup::Text)
	group_ents.erase_entities(edges)
	group_ents.erase_entities(text)
	end

	##
	def get_layer_by_name(layer_name)
	layers=Sketchup.active_model.layers
	layer=layers[layer_name]
	if layer
		return layer
	else
		return layers[0]
	end
	end
	##
	def get_material()
	mat_name=self.material_name
	mats=Sketchup.active_model.materials
	if mat_name
		mat=mats[mat_name]
		if mat
			return mat
		end
	end
	return nil
	end
	###
	def fix_vector!(vec)
	x=vec.x
	y=vec.y
	z=vec.z
	x=x.abs if x==0.0
	y=y.abs if y==0.0
	z=z.abs if z==0.0
	vec.set!(x,y,z)
	end
	###

	####Exceptions
	class TrimException < RuntimeError
	end
	
	
	##
	def get_cap_curve_vertices(start)
	ents=self.entities
	edges=ents.grep(Sketchup::Edge)
	verts=Sketchup::Set.new
	profile=self.profile
	num_points=profile.points.length
	curve_edges=edges.find_all {|edge| edge.curve!=nil}
	curves=curve_edges.collect {|edge| edge.curve}
	curves.uniq!
	
	if start
		cap_normal=start_extend_vector()
		point=self.chain[0]
	else
		cap_normal=end_extend_vector()
		point=self.chain[-1]
	end
	curves=curves.find_all {|curve| curve.vertices.length==num_points}
	if curves.length>1
		cap_curve=get_closest_face_to_point(curves,point)
	else
		cap_curve=curves[0]
	end
	return cap_curve.vertices
	end
	######################
	def soften(angle,harden=false)  #softens edges of the profile member based on the given angle (degrees)
	return if angle==0.0
	angle=angle.degrees
	ents=self.entities   #get the entities inside the profile member
	edges=ents.grep(Sketchup::Edge)   #get all the edges inside the profile member
	num_edges=edges.length.to_f
	i=0.0
	last_update=Time.now
	edges.each {|edge|
		t=Time.now
		faces=edge.faces
		if faces.length==2
			n1=faces[0].normal
			n2=faces[1].normal
			angle2=n1.angle_between(n2)
			if angle2<=angle and harden==false
				edge.soft=true
				edge.smooth=true
			elsif angle2>=angle and harden==true
				edge.soft=false
				edge.smooth=false
			end
		end
		i+=1
		if (t-last_update)>0.5
			Sketchup.set_status_text("Smoothing Edges - #{((i/num_edges)*100).to_i}%")
			last_update=Time.now
		end
	}
	return nil
	end
	#
	def validate_for_trim(inst)
	if @group.class==Sketchup::Group
		@group=@group.make_unique
		@su_defn=@group.definition
	end
	pm=ProfileMember.new(inst)
	if pm.definition==self.definition
		inst=inst.make_unique
	end
	return inst
	end
	##
	def prepare_end_for_trim(trimmer_inst,cap_index)
		begin
			path=self.chain
			cf=get_cap_faces()[cap_index]
			push_length=push_cap_face(cf)
			max_distances_to_entity=[]
			max_distance_to_entity=get_max_perp_distance_from_face_to_entity(cf,trimmer_inst)
			pull_length=pull_cap_face(cf,max_distance_to_entity,push_length)
			end_offset_distance=pull_length-push_length
			offset_chain_end(cap_index,end_offset_distance)	
		end
		
	end
	###
	def push_cap_face(cf)
	range=get_range_perpindicular_edge_length(cf.outer_loop.vertices)
	push_length=range.min
	push_length+=-0.01
	cf.pushpull(-push_length)
	
	return push_length
	end
	##
	def pull_cap_face(cap_face,distance,push_length)
	if distance==nil
		raise TrimException, "不能修剪/延伸.请确保选择的两个轮廓放样的模型路径相交.", caller
	end
		
	if distance
		normal=cap_face.normal
		normal.length=distance+push_length
		cap_face.pushpull(distance)
	end
	
	return distance
	
	end
	
	##
	def offset_chain_end(cap_index,distance)
	
	path=self.chain
	if cap_index==0
		vec=start_extend_vector()
		pt_index=0
	else
		vec=end_extend_vector()
		pt_index=-1
	end
	vec.length=distance
	path[pt_index]=path[pt_index].offset(vec)
	self.set_chain(path)
	
	end
	###
	def perform_boolean_trim(trimmer_inst)
	name=@group.name
	instances=self.definition.instances
	defn_name=@su_defn.name
	dicts_hash={}
	dicts=@su_defn.attribute_dictionaries
	dicts.each {|dict|
		dict_name=dict.name
		dict_hash={}
		dict.each {|key,val| dict_hash[key]=val}
		dicts_hash[dict_name]=dict_hash
	}
	result=trimmer_inst.trim(self.instance)
	result.name=name
	defn=result.entities.parent
	dicts_hash.each {|dict_name,dict|
		dict.each {|key,val| defn.set_attribute(dict_name,key,val)}
	}
	if instances.length>1
		@group=result.to_component
		@group.name=name
		@su_defn=defn
		@su_defn.name=defn_name
		instances.each {|inst| inst.definition=@su_defn if inst.valid?}			
	end
	return defn
	end
	#
	def delete_waste(defn)
	ents=defn.entities
	faces=ents.grep(Sketchup::Face)
	connected_faces=[]
	connected_area_sum=[]
	area_sum=0
	faces.each_index {|i|
		area_sum=0
		connected_faces=faces[i].all_connected
		connected_faces=connected_faces.grep(Sketchup::Face)
		connected_faces.each {|f| area_sum+=f.area}
		connected_area_sum[i]=area_sum
	}
	uniq_area_sum=connected_area_sum.uniq
	while ((uniq_area_sum.min)!=(uniq_area_sum.max))
		min_area=uniq_area_sum.min
		face_index=connected_area_sum.index(min_area)
		reference_face=faces[face_index]
		connected_faces=reference_face.all_connected
		ents.erase_entities(connected_faces)
		uniq_area_sum.delete(min_area)
	end
		
	end
	##
	def project_vertices_to_plane(verts,plane)
	pm_trans=self.transformation
	normal=get_normal_of_vertices(verts).transform(pm_trans)
	vectors=[]
	verts.each {|v|
		p1=v.position.transform(pm_trans)
		line=[p1,normal]
		p2=Geom.intersect_line_plane(line,plane)
		vectors.push(p1.vector_to(p2))
	}
	vectors=vectors.collect {|vec| vec.transform(pm_trans.inverse)}
	if self.profile.is_1d?
		curve_edge=verts[0].edges.find {|edge| edge.curve!=nil}
		curve=curve_edge.curve
		new_positions=[]
		vectors.each_index {|i| new_positions[i]=verts[i].position.transform(Geom::Transformation.new(vectors[i]))}
		curve.move_vertices(new_positions)
	else
		common_face=(verts[0].faces)&(verts[1].faces)&(verts[2].faces)
		common_face=common_face[0]
		if common_face
			outer_loop_verts=common_face.outer_loop.vertices
			loops=[]
			face_loops=common_face.loops
			face_loops.each_index {|i|
				loop_verts=face_loops[i].vertices
				loops<<face_loops[i].vertices unless face_loops[i].outer?
			}
			ents=self.entities
			ents.erase_entities(common_face)
			
			new_face_pts=[]
			vectors.each_index {|i| new_face_pts[i]=verts[i].position.transform(Geom::Transformation.new(vectors[i]))}
			ents.transform_by_vectors(verts,vectors)
			ents.add_face(new_face_pts)
			
			loops.each {|loop_verts|
				loop_vectors=[]
				loop_verts.each {|v|
					p1=v.position.transform(pm_trans)
					line=[p1,normal]
					p2=Geom.intersect_line_plane(line,plane)
					loop_vectors.push(p1.vector_to(p2))
				}
				
				loop_vectors=loop_vectors.collect {|vec| vec.transform(pm_trans.inverse)}
				new_loop_pts=[]
				loop_vectors.each_index {|i| new_loop_pts[i]=loop_verts[i].position.transform(Geom::Transformation.new(loop_vectors[i]))}
				ents.transform_by_vectors(loop_verts,loop_vectors)
				hole_face=ents.add_face(new_loop_pts)
				hole_face.erase!
			}
		end
	end
	end
	##
	def get_closest_cap_verts_to_point(point)
	cap_vertices=self.get_cap_vertices()
	trans=self.transformation
	distance_to_point=[]
	cap_vertices.delete(nil)
	if cap_vertices.empty?
		raise TrimException, "不能修剪/延伸.轮廓放样模型必须有一个未修改的端面.", caller
	end
	cap_vertices.each_index {|index|
		verts=cap_vertices[index]
		box=Geom::BoundingBox.new
		face_pts=verts.collect {|v| v.position.transform(trans)}
		box.add(face_pts)
		face_center=box.center
		distance_to_point.push(face_center.vector_to(point).length)
	}
	min_distance=distance_to_point.min
	min_index=distance_to_point.index(min_distance)
	return cap_vertices[min_index]
	end

end #class

end #module PB
end #module DM
